/*
   Coolpix RAW to Coolpix NEF converter
   by paul69, v0.02, 2003-01-22, <e2500@narod.ru>

   Supported camera (tested)
      E2500
      E4500
      E995
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define datetime __DATE__ " " __TIME__
#define version "v0.02"
#define program "raw2nef"
#define author "paul69 <e2500@narod.ru>"

class TFile
{
public:
   TFile(FILE* p=NULL) : fp(p) {}
   ~TFile() { if(fp) fclose(fp); }

   FILE* fp;
   operator FILE* () { return fp; }

   void operator = (FILE* p) { if(fp) fclose(fp); fp = p; }
   void close() { if(fp) fclose(fp); fp=NULL; }
};

   enum { Gr, Mg, Cy, Ye };
   enum { R, G, B };

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;
typedef unsigned short color_t;

class TBuff
{
public:
   TBuff() : p(NULL) {}
   TBuff(unsigned sz) : p(new color_t[sz]) {}
   ~TBuff() { delete [] p; }

   color_t* p;
   operator color_t* () { return p; }
   color_t& operator [] (int i) { return p[i]; }

   bool alloc(unsigned sz) { delete [] p; return NULL != (p = new color_t[sz]); }
   void free() { delete [] p; p = NULL; }
};

////////////////////////////////////////////////////////////////////////////////
struct TCCDField
{
   unsigned offset, rows;
};

struct TCCDParam
{
   unsigned raw_width;
   unsigned raw_height;
   unsigned file_size;

   char cfa_pattern[4];

   TCCDField field[2];
};


bool read_ccd(FILE* in, color_t ccd[], TCCDParam& ccd_pars)
{
   unsigned ccd_size = ccd_pars.raw_width * ccd_pars.raw_height;

   for(int f=0; f<2; ++f)
   {
      if( ccd_pars.field[f].rows == 0 ) continue;

      fseek( in, ccd_pars.field[f].offset, SEEK_SET );

      unsigned raw_width = ccd_pars.raw_width;
      unsigned line = f * raw_width;
      unsigned delta = 2 * raw_width;
      unsigned field_size = ccd_pars.field[f].rows * raw_width*12/8;

      unsigned i, x;

      for(x=i=0; i<field_size; i += 3)
      {
         byte bitbuf[3];
         if( !fread( bitbuf, 3, 1, in) )
            return false;

         if( x == raw_width )
         {
            line += delta;
            x = 0;

            // check space for row
            if( line + raw_width > ccd_size )
               return false;
         }

         ccd[line + x] = (color_t)(((bitbuf[0]<<4)|((bitbuf[1]>>4)&15)));
         ++x;

         ccd[line + x] = (color_t)((((bitbuf[1]&15)<<8)|(bitbuf[2])));
         ++x;
      }
   }
   return true;
}

////////////////////////////////////////////////////////////////////////////////

struct TIFF
{
   enum type_of_data_in_TIFF
   {
      type_BYTE = 1, type_ASCII, type_WORD, type_DWORD, type_RATIONAL,
      type_SBYTE, type_UNDEFINED, type_SHORT, type_LONG, type_SRATIONAL,
      type_FLOAT, /* type_DOUBLE, - unimplemented */
      n_types
   };

   enum used_tags
   {
      // standard TIFF tags
      NewSubfile = 0x00FE,
      ImageWidth = 0x0100,
      ImageHeight = 0x0101,
      BitsPerSample = 0x0102,
      Compression = 0x0103,
      PhotometricInterpretation = 0x0106,
      ImageDescription = 0x010E,
      Make = 0x010F,
      Model = 0x0110,
      StripOffsets = 0x0111,
      Orientation = 0x0112,
      SamplesPerPixel = 0x0115,
      RowsPerStrip = 0x0116,
      StripByteCounts = 0x0117,
      XResolution = 0x011A,
      YResolution = 0x011B,
      PlanarConfiguration = 0x011C,
      ResolutionUnit = 0x0128,
      Software = 0x0131,
      DateTime = 0x0132,
      SubIDFs = 0x14A,
      ReferenceBlackWhite = 0x0214,
      DateTimeOriginal = 0x9003,
      DateTimeDigitized = 0x9004,
      Copyright = 0x8298,
      // nikon specific tags
      CFA_size = 0x828D,
      CFA_pattern = 0x828E,

      ExposureTime = 0x829A,
      F_Number = 0x829D,
      ExposureProgram = 0x8822,
      ExposureBias = 0x9204,
      MaxApertureValue = 0x9205,
      MeteringMode = 0x9207,
      FocalLength = 0x920A,
      MakerNote = 0x927C,
      UserComment = 0x9286,
      SubSecTime = 0x9290,
      SubSecTimeOriginal = 0x9291,
      SubSecTimeDigitized = 0x9292,
      SensingMethod = 0xA217,
      FileSource = 0xA300,
      SceneType = 0xA301,
      ColorFilterArrayPattern = 0xA302,
   };

   static int sizeof_value[n_types];
};

int TIFF::sizeof_value[n_types] = {0,1,1,2,4,8,1,1,2,4,8,4};

struct rational { dword n, d; };
struct Rational : rational { Rational(int a, int b) {n=a; d=b;} };

class IFDir
{
public:
   IFDir() : tag_count_(0), array_(NULL), arr_size_(0) {}
   ~IFDir();

   struct Tag
   {
      int tag;
      int type;
      int count;
      char* value;

      Tag(int _tag, int _type, int _count, const void* _value);
      ~Tag() { delete [] value; }

      int get_size() { return count * TIFF::sizeof_value[type]; }
   };

   bool add(int tag, int type, int count, const void* value);
   bool add_WORD(int tag, word value) { return add(tag, TIFF::type_WORD, 1, &value); }
   bool add_DWORD(int tag, dword value) { return add(tag, TIFF::type_DWORD, 1, &value); }
   bool add_RATIONAL(int tag, rational value) { return add(tag, TIFF::type_RATIONAL, 1, &value); }
   bool add_ASCII(int tag, char* value) { return add(tag, TIFF::type_ASCII, strlen(value)+1, value); }
   bool add_ASCII(int tag, char* value, int len) { return add(tag, TIFF::type_ASCII, len, value); }
   unsigned get_size();
   unsigned get_tag_offset(word tag);
   Tag* get_tag(word tag);
   bool write_to_buffer( char* buf, int size, bool endian, unsigned idf_offset, unsigned next_ifd);
   bool write(FILE* out, bool endian, unsigned idf_offset, unsigned next_ifd);
public:
   int tag_count_;
   Tag** array_;
   int arr_size_;
};

IFDir::~IFDir()
{
   for(int i=0; i<tag_count_; ++i)
      { delete array_[i]; }
   delete [] array_;
}

IFDir::Tag::Tag(int _tag, int _type, int _count, const void* _value)
   : tag(_tag), type(_type), count(_count), value(NULL)
{
   int size = get_size();
   value = new char[size];
   
   if( _value )
      memcpy( value, _value, size );
   else
      memset( value, 0, size );
}

bool IFDir::add(int tag, int type, int count, const void* value)
{
   if( type >= TIFF::n_types || count == 0 )
      return false;

   int i;
   for(i=0; i<tag_count_; ++i)
   {
      if( array_[i]->tag == tag )
      {
         delete array_[i];
         array_[i] = new Tag(tag, type, count, value);
         return true;
      }
      if( array_[i]->tag > tag )
         break;
   }

   if( tag_count_ == arr_size_ )
   {
      int new_size = arr_size_ + 64;
      Tag** temp = array_;
      array_ = new Tag*[new_size];

      if( !array_ )
      {
         array_ = temp;
         return false;
      }

      arr_size_ = new_size;

      if( tag_count_ )
         memcpy( array_, temp, sizeof(array_[0])*tag_count_ );

      delete [] temp;
   }

   if( i < tag_count_ )
   {
      memmove( array_+i+1, array_+i, sizeof(array_[0])*(tag_count_-i) );
   }

   array_[i] = new Tag(tag, type, count, value);

   ++tag_count_;
   return true;
}

unsigned IFDir::get_size()
{
   unsigned size = 2 + 12*tag_count_ + 4;
   for(int i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];
      unsigned tag_data_size = t->get_size();
      if( tag_data_size > 4 )
         size += tag_data_size;
   }
   return size;
}

IFDir::Tag* IFDir::get_tag(word tag)
{
   for(int i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];

      if( t->tag == tag )
         return t;
   }
   return NULL;
}

unsigned IFDir::get_tag_offset(word tag)
{
   unsigned offset = 2 + 12*tag_count_ + 4;
   for(int i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];
      unsigned tag_data_size = t->get_size();

      if( t->tag == tag )
      {
         if( tag_data_size > 4 )
            return offset;

         return 2 + 12*i + 2+2+4;
      }
      if( tag_data_size > 4 )
         offset += tag_data_size;
   }
   return 0;
}

bool IFDir::write(FILE* out, bool endian, unsigned tiff_start, unsigned next_ifd)
{
   unsigned size = get_size();
   char* buffer = new char[size];

   bool ret = false;

   unsigned idf_start = ftell(out) - tiff_start;

   if( write_to_buffer( buffer, size, endian, idf_start, next_ifd) )
   {
      ret = fwrite( buffer, size, 1, out) == 1;
   }
   delete [] buffer;
   return ret;
}

bool write_word(FILE* out, bool endian, unsigned w)
{
   if( endian )
   {
      if( fputc( (int)(w & 0xFF), out) == EOF ) return false;
      return fputc( (int)((w >> 8)&0xFF), out) != EOF;
   }

   if( fputc( (int)((w >> 8)&0xFF), out) == EOF ) return false;
   return fputc( (int)(w & 0xFF), out) != EOF;
}

bool write_dword(FILE* out, bool endian, unsigned dw)
{
   if( endian )
   {
      if( fputc( (int)((dw) & 0xFF), out) == EOF ) return false;
      if( fputc( (int)((dw >> 8) & 0xFF), out) == EOF ) return false;
      if( fputc( (int)((dw >> 16) & 0xFF), out) == EOF ) return false;
      return fputc( (int)((dw >> 24) & 0xFF), out) != EOF;
   }

   if( fputc( (int)((dw >> 24) & 0xFF), out) == EOF ) return false;
   if( fputc( (int)((dw >> 16) & 0xFF), out) == EOF ) return false;
   if( fputc( (int)((dw >> 8) & 0xFF), out) == EOF ) return false;
   return fputc( (int)((dw) & 0xFF), out) != EOF;
}

class TOut
{
public:
   TOut( char* buf, int size, bool endian ) : buf_(buf), size_(size), off_(0), endian_(endian) {}

   char* buf_;
   int size_;
   int off_;
   bool endian_;

public:
   void put(int c)
   {
      if( off_ < size_ )
      {
         buf_[off_++] = (char)c;
      }
   }

   bool is_place() { return off_ < size_; }
   bool is_place(int n) { return off_+n <= size_; }

   bool write_byte(unsigned v);
   bool write_word(unsigned w);
   bool write_dword(unsigned dw);
};

bool TOut::write_byte(unsigned v)
{
   if( !is_place(1) ) return false;

   put((int)v);

   return  true;
}

bool TOut::write_word(unsigned w)
{
   if( !is_place(2) ) return false;

   if( endian_ )
   {
      put( (int)(w & 0xFF) );
      put( (int)((w >> 8)&0xFF) );
      return true;
   }

   put( (int)((w >> 8)&0xFF) );
   put( (int)(w & 0xFF) );
   return true;
}

bool TOut::write_dword(unsigned dw)
{
   if( !is_place(4) ) return false;

   if( endian_ )
   {
      put( (int)((dw) & 0xFF) );
      put( (int)((dw >> 8) & 0xFF) );
      put( (int)((dw >> 16) & 0xFF) );
      put( (int)((dw >> 24) & 0xFF) );
      return true;
   }

   put( (int)((dw >> 24) & 0xFF) );
   put( (int)((dw >> 16) & 0xFF) );
   put( (int)((dw >> 8) & 0xFF) );
   put( (int)((dw) & 0xFF) );
   return true;
}

bool IFDir::write_to_buffer( char* buf, int size, bool endian, unsigned idf_start, unsigned next_ifd)
{
   TOut out( buf, size, endian );
   if( !out.write_word(tag_count_) )
      return false;

   int i, k;
   unsigned data_offset = idf_start + 2 + 12*tag_count_ + 4;

   for(i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];

      out.write_word(t->tag);
      out.write_word(t->type);
      out.write_dword(t->count);

      unsigned tag_data_size = t->get_size();
      if( tag_data_size > 4 )
      {
         out.write_dword(data_offset);
         data_offset += tag_data_size;
      }
      else
      {
         switch(t->type)
         {
         case TIFF::type_BYTE:
         case TIFF::type_SBYTE:
         case TIFF::type_ASCII:
         case TIFF::type_UNDEFINED:
            for(k=0; k<4; ++k)
            {
               out.write_byte(k < t->count ? ((byte*)t->value)[k] : (byte)0 );
            }
            break;

         case TIFF::type_WORD:
         case TIFF::type_SHORT:
            for(k=0; k<2; ++k)
            {
               out.write_word(k < t->count ? ((word*)t->value)[k] : (word)0 );
            }
            break;

         case TIFF::type_DWORD:
         case TIFF::type_LONG:
         case TIFF::type_FLOAT:
            out.write_dword(*(dword*)t->value);
            break;
         }
      }
   }

   out.write_dword(next_ifd);

   for(i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];

      unsigned tag_data_size = t->get_size();
      if( tag_data_size > 4 )
      {
         switch(t->type)
         {
         case TIFF::type_BYTE:
         case TIFF::type_SBYTE:
         case TIFF::type_ASCII:
         case TIFF::type_UNDEFINED:
            for(k=0; k<t->count; ++k)
            {
               out.write_byte(((byte*)t->value)[k]);
            }
            break;

         case TIFF::type_WORD:
         case TIFF::type_SHORT:
            for(k=0; k<t->count; ++k)
            {
               out.write_word(((word*)t->value)[k]);
            }
            break;

         case TIFF::type_DWORD:
         case TIFF::type_LONG:
         case TIFF::type_FLOAT:
            for(k=0; k<t->count; ++k)
            {
               out.write_dword(((dword*)t->value)[k]);
            }
            break;

         case TIFF::type_RATIONAL:
         case TIFF::type_SRATIONAL:
            for(k=0; k<t->count; ++k)
            {
               out.write_dword(((rational*)t->value)[k].n);
               out.write_dword(((rational*)t->value)[k].d);
            }
            break;
         }
      }
   }
   return true;
}

byte some_data[1280] =
{
0x00,0x00,0x00,0x02,0x00,0x00,0x00,0x48,0x00,0x00,0x0F,0xFF,0x00,0x00,0x00,0x03
,0x76,0x76,0x89,0x8B,0x00,0x00,0x00,0xFF,0x00,0x01,0x01,0xC2,0x00,0x00,0x1F,0xFF
,0xFF,0xFF,0xFF,0xFF,0x1C,0x18,0x11,0x18,0x14,0x15,0x15,0x17,0x00,0x00,0x00,0x00
,0x00,0x00,0x01,0xFF,0x00,0x00,0x07,0xFF,0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x00
,0x00,0x00,0x00,0x94,0x00,0x00,0x01,0x20,0x00,0x00,0x00,0x14,0x00,0x00,0x00,0x8B
,0x00,0x00,0x00,0x40,0x00,0x00,0x00,0x40,0x00,0x00,0x00,0x8F,0x00,0x00,0x03,0xAF
,0x00,0x00,0x00,0xD7,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x02,0x60,0x4E
,0x01,0x02,0x10,0x41,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xE4
,0x00,0x00,0x02,0x5A,0x00,0x00,0x00,0xC6,0x00,0x58,0x00,0xE7,0x00,0x06,0x00,0x54
,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x0C,0x00,0x0B
,0x00,0x0A,0x00,0x0F,0x00,0x00,0x00,0x14,0x00,0x00,0x00,0x04,0x00,0x02,0x00,0xBE
,0x00,0x02,0x03,0x3E,0x00,0x00,0x01,0xFF,0x00,0x40,0x00,0x4D,0x00,0x00,0x00,0x00
,0x00,0x03,0x03,0x52,0x00,0x00,0x00,0x0F,0x00,0x00,0x00,0x01,0x00,0x00,0x00,0x00
,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
,0x00,0x4E,0x00,0x19,0x00,0x41,0x00,0x1E,0x00,0x62,0x00,0x6E,0x00,0x41,0x00,0x1E
,0x00,0x39,0x00,0x05,0x00,0x41,0x00,0x1E,0x00,0x66,0x00,0x05,0x00,0x41,0x00,0x1E
,0x00,0x36,0x00,0x12,0x00,0x41,0x00,0x1E,0x00,0x6E,0x00,0x7E,0x00,0x6E,0x00,0x7E
,0x00,0x6E,0x00,0x7F,0x00,0x7D,0x00,0x71,0x00,0x78,0x00,0x75,0x00,0x73,0x00,0x7A
,0x00,0x69,0x00,0x84,0x00,0x64,0x00,0x89,0x00,0x5E,0x00,0x8D,0x00,0x57,0x00,0xA8
,0x00,0x62,0x00,0x95,0x00,0x5E,0x00,0x9B,0x00,0x5A,0x00,0xA2,0x00,0x53,0x00,0xAE
,0x00,0x4F,0x00,0xB5,0x00,0x4A,0x00,0xBB,0x00,0x5F,0x00,0xDC,0x00,0x79,0x00,0x97
,0x00,0xA2,0x00,0x76,0x00,0x97,0x00,0x71,0x00,0xAB,0x00,0x64,0x00,0xA5,0x00,0x68
,0x00,0x9E,0x00,0x6D,0x00,0x90,0x00,0x71,0x00,0x89,0x00,0x7A,0x00,0x81,0x00,0x7E
,0x00,0x98,0x00,0x70,0x00,0xB7,0x00,0x5D,0x00,0xAC,0x00,0x63,0x00,0xA2,0x00,0x69
,0x00,0x8D,0x00,0x76,0x00,0x82,0x00,0x7C,0x00,0x77,0x00,0x82,0x00,0x0B,0x00,0x0C
,0x00,0x0F,0x00,0x0A,0x00,0x00,0x00,0x14,0x00,0x06,0x00,0x0C,0x00,0x09,0x00,0x0A
,0x00,0x00,0x00,0x0F,0x00,0x10,0x00,0x0C,0x00,0x15,0x00,0x0A,0x00,0x00,0x00,0x19
,0x00,0x01,0x00,0x0C,0x00,0x01,0x00,0x0A,0x00,0x00,0x00,0x05,0x00,0x00,0x00,0x02
,0x00,0x00,0x03,0x3E,0x00,0x00,0x00,0x0F,0x00,0x01,0x01,0xC2,0x76,0x76,0x89,0x8B
,0x1C,0x18,0x11,0x18,0x14,0x15,0x15,0x17,0x00,0x00,0x00,0x94,0x00,0x00,0x01,0x20
,0x00,0x00,0x00,0x14,0x00,0x00,0x00,0x8B,0x00,0x00,0x00,0x40,0x00,0x00,0x00,0x40
,0x00,0x00,0x00,0x8F,0x00,0x00,0x03,0xAF,0x00,0x00,0x00,0xD7,0x00,0x00,0x00,0xE7
,0x00,0x00,0x00,0x58,0x00,0x00,0x00,0x54,0x00,0x00,0x00,0x06,0x00,0x00,0x00,0x03
,0x00,0x00,0x03,0x48,0x00,0x00,0x00,0x08,0x00,0x03,0x02,0x8A,0x90,0x60,0x93,0x82
,0x1B,0x12,0x11,0x18,0x1D,0x16,0x19,0x16,0x00,0x00,0x00,0x94,0x00,0x00,0x01,0x08
,0x00,0x00,0x00,0x27,0x00,0x00,0x00,0x90,0x00,0x00,0x00,0x40,0x00,0x00,0x00,0x40
,0x00,0x00,0x00,0x8F,0x00,0x00,0x03,0xAF,0x00,0x00,0x00,0xCB,0x00,0x00,0x00,0xEC
,0x00,0x00,0x00,0x66,0x00,0x00,0x00,0x54,0x00,0x00,0x00,0xF6,0x00,0x00,0x00,0x4E
,0x00,0x00,0x00,0x45,0x00,0x00,0x00,0x4E,0x00,0x00,0x00,0x47,0x00,0x00,0x00,0x00
,0x00,0x00,0x00,0xE3,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFC,0x00,0x00,0x00,0xE3
,0x00,0x00,0x00,0x80,0x00,0x00,0x00,0xFC,0x00,0x00,0x00,0x0A,0x00,0x00,0x00,0x00
,0x00,0x00,0x00,0x05,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x06,0x1F
,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
,0x00,0x00,0x00,0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x07,0xCF,0x00,0x00,0x03,0xD9
,0x00,0x00,0x0F,0x72,0x00,0x00,0x05,0x9E,0x00,0x00,0x03,0xE3,0x00,0x00,0x00,0xE1
};

#define tn_H 120
#define tn_W 160

bool ccd_interpolation(color_t ccd[], color_t image[], TCCDParam& ccd_pars)
{
   int cfa[4]; // index of color filter

   for(int i=0; i<4; ++i)
   {
      switch( ccd_pars.cfa_pattern[i] )
      {
      case 'G': cfa[i] = Gr; break;
      case 'M': cfa[i] = Mg; break;
      case 'C': cfa[i] = Cy; break;
      case 'Y': cfa[i] = Ye; break;
      default: return false;
      }
   }

   int height = ccd_pars.raw_height;
   int width = ccd_pars.raw_width;

   int dx = width/tn_W;
   int dy = height/tn_H;

   color_t* out = image;

   for(int y=0; y<tn_H; ++y)
   {
      int _y = y * dy;
      for(int x=0; x<tn_W; ++x, out += 3)
      {
         int c[4];

         int _x = x * dx;
         color_t* in = &ccd[_y*width + _x];
         // index in CFA
         int i = (_x&1) | ((_y&1)<<1);

         c[ cfa[i] ] = in[0];
         c[ cfa[i^1] ] = in[1];
         c[ cfa[i^2] ] = in[width];
         c[ cfa[i^3] ] = in[width+1];

         float YY = (c[Gr] + c[Mg] + c[Ye] + c[Cy])/4;
         float RY = (c[Mg] + c[Ye]) - (c[Gr] + c[Cy]);
         float BY = (c[Mg] + c[Cy]) - (c[Gr] + c[Ye]);

         int RR = (int)(0.4f*YY + 0.4f*RY - 0.1f*BY);
            if( RR < 0 ) RR=0; if( RR > 65535 ) RR = 65535;

         int GG = (int)(0.8f*YY - 0.2f*RY - 0.2f*BY);
            if( GG < 0 ) GG=0; if( GG > 65535 ) GG = 65535;

         int BB = (int)(0.4f*YY - 0.1f*RY + 0.4f*BY);
            if( BB < 0 ) BB=0; if( BB > 65535 ) BB = 65535;

         out[ R ] = (color_t)RR;
         out[ G ] = (color_t)GG;
         out[ B ] = (color_t)BB;
      }
   }
   return true;
}

bool write_nef(FILE* out, bool endian, color_t ccd[], TCCDParam& ccd_pars)
{
   int i;
   IFDir IFD0;

   unsigned ifd0_start = 8;
   unsigned ifd0_data_size = tn_W*tn_H*3; //57600;
   IFD0.add_DWORD( TIFF::NewSubfile, 1 ); // thumbnail
   IFD0.add_DWORD( TIFF::ImageWidth, tn_W );
   IFD0.add_DWORD( TIFF::ImageHeight, tn_H );
   const int samples_per_pixel = 3;
   word bits_per_sample[samples_per_pixel] = {8,8,8};
   IFD0.add( TIFF::BitsPerSample, TIFF::type_WORD, samples_per_pixel, bits_per_sample );
   IFD0.add_WORD( TIFF::Compression, 1 );
   IFD0.add_WORD( TIFF::PhotometricInterpretation, 2 ); // RGB
   IFD0.add_ASCII( TIFF::ImageDescription, "           " );
   IFD0.add_ASCII( TIFF::Software, "E5700v1.0" );
   IFD0.add_ASCII( TIFF::Make, "NIKON" );
   IFD0.add_ASCII( TIFF::Model, "E5700" );
   IFD0.add_WORD( TIFF::Orientation, 1 ); //  TopLeft
   IFD0.add_DWORD( TIFF::StripOffsets, 0 );
   IFD0.add_WORD( TIFF::SamplesPerPixel, samples_per_pixel );
   IFD0.add_DWORD( TIFF::RowsPerStrip, tn_H);
   IFD0.add_DWORD( TIFF::StripByteCounts, ifd0_data_size );

   IFD0.add_RATIONAL( TIFF::XResolution, Rational(300,1) );
   IFD0.add_RATIONAL( TIFF::YResolution, Rational(300,1) );
   IFD0.add_WORD( TIFF::PlanarConfiguration, 1 ); // chunky
   IFD0.add_WORD( TIFF::ResolutionUnit, 2 ); // inch

   dword sub_ifd = 0;
   IFD0.add_DWORD( TIFF::SubIDFs, sub_ifd ); // SubIDF's offsets
   IFD0.add_DWORD( 0x8769, 0 ); // EXIF offset

   unsigned ifd0_size = IFD0.get_size();
   unsigned exif_start = ifd0_start+ifd0_size;
   IFD0.add_DWORD( 0x8769, exif_start ); // EXIF offset

   IFDir EXIF;

   IFDir MakerNote;
   MakerNote.add ( 0x0014, TIFF::type_UNDEFINED, 1280, some_data );

   // reserve place
   EXIF.add( TIFF::MakerNote, TIFF::type_UNDEFINED, MakerNote.get_size() + 18, NULL );

   byte cfa_pattern[8] = {0,2,0,2,5,3,1,4}; // MGCY
   EXIF.add( TIFF::ColorFilterArrayPattern, TIFF::type_UNDEFINED, 8, cfa_pattern );

   IFDir::Tag* maker_note = EXIF.get_tag( TIFF::MakerNote );
   if( maker_note )
   {
      memcpy( maker_note->value, "Nikon\0\002\0\0\0", 10 );
      memcpy( maker_note->value+10, "II*\0\010\0\0\0", 8 );
      MakerNote.write_to_buffer( maker_note->value+18, maker_note->get_size(), true, 8, 0 );
   }

   unsigned exif_size = EXIF.get_size();

   sub_ifd = exif_start + exif_size;
   IFD0.add_DWORD( TIFF::SubIDFs, sub_ifd );

   IFDir IFD;

   int image_W = 2576;
   int image_H = 1924;

   IFD.add_DWORD( TIFF::NewSubfile, 0 );
   IFD.add_DWORD( TIFF::ImageWidth, image_W );
   IFD.add_DWORD( TIFF::ImageHeight, image_H );
   IFD.add_WORD( TIFF::BitsPerSample, 12 );
   IFD.add_WORD( TIFF::Compression, 1 );
   IFD.add_WORD( TIFF::PhotometricInterpretation, 0x8023 ); // nikon specific

   unsigned row_size = (image_W*12+7)/8;
   unsigned image_size = image_H * row_size;

   IFD.add_DWORD( TIFF::StripOffsets, 0 );
   IFD.add_WORD( TIFF::Orientation, 1 ); //  TopLeft
   IFD.add_WORD( TIFF::SamplesPerPixel, 1 );
   IFD.add_DWORD( TIFF::RowsPerStrip, image_H);
   IFD.add_DWORD( TIFF::StripByteCounts, image_size );

   IFD.add_RATIONAL( TIFF::XResolution, Rational(300,1) );
   IFD.add_RATIONAL( TIFF::YResolution, Rational(300,1) );
   IFD.add_WORD( TIFF::PlanarConfiguration, 1 ); // chunky
   IFD.add_WORD( TIFF::ResolutionUnit, 2 ); // inch

   // nikon specific tags
   word cfa_size[2] = { 2, 2 };
   byte cfa_patt[4] = { 5, 3, 1, 4 };

   IFD.add( TIFF::CFA_size, TIFF::type_WORD, 2, cfa_size );
   IFD.add( TIFF::CFA_pattern, TIFF::type_BYTE, 4, cfa_patt );

   unsigned ifd_size = IFD.get_size();
   unsigned image0_offset = sub_ifd + ifd_size;
   unsigned image_offset = image0_offset + ifd0_data_size;
   IFD0.add_DWORD( TIFF::StripOffsets, image0_offset );
   IFD.add_DWORD( TIFF::StripOffsets, image_offset );

   // writing
   if( !write_word(out, endian, *(word*)(endian ? "II" : "MM")) )
      return false;

   write_word(out, endian, '*');
   write_dword(out, endian, 8);

   if( !IFD0.write(out, endian, 0, 0) )
      return false;

   if( ftell(out) != exif_start )
   {
      printf("error: exif_start=%08X curr=%08X\n", exif_start, ftell(out));
      return false;
   }

   if( !EXIF.write(out, endian, 0, 0) )
      return false;

   if( ftell(out) != sub_ifd )
   {
      printf("error: idf2_start=%08X curr=%08X\n", sub_ifd, ftell(out));
      return false;
   }

   if( !IFD.write(out, endian, 0, 0) )
      return false;

   if( ftell(out) != image0_offset )
   {
      printf("error: img1_start=%08X curr=%08X\n", image0_offset, ftell(out));
      return false;
   }

   // write tiff color thumbnail
   {
      TBuff tn(tn_W*tn_H*3);

      ccd_interpolation( ccd, tn, ccd_pars);

      int max_color[3];
      max_color[0]=max_color[1]=max_color[2]=0;

      for(i=0; i<tn_W*tn_H; ++i)
      {
         for(int j=0; j<3; ++j)
         {
            int pixel = tn[i*3+j];
            if( pixel > max_color[j] )
               max_color[j] = pixel;
         }
      }

      for(i=0; i<tn_W*tn_H; ++i)
      {
         for(int j=0; j<3; ++j)
         {
            int pixel = tn[i*3+j];

            pixel = max_color[j] ? (int)(pixel*300/max_color[j]) : pixel>>4;

            if( pixel > 255 ) pixel = 255;
            fputc(pixel, out);
         }
      }
   }

   if( ftell(out) != image_offset )
   {
      printf("error: image_start=%08X curr=%08X\n", image_offset, ftell(out));
      return false;
   }

   bool ret = true;
   byte* row = new byte[row_size];

   memset(row, 0, sizeof(row[0])*row_size );

   int first = 0; // even rows to first field
   int second = 1; // odd rows to second field
   bool pix_shift = false;

   if( *(long*)ccd_pars.cfa_pattern == *(long*)"YCGM"
    || *(long*)ccd_pars.cfa_pattern == *(long*)"CYMG" )
   {
      first = -1; // shift picture down, first field has odd rows
      second = 0; // second field has even rows
   }

   if( *(long*)ccd_pars.cfa_pattern == *(long*)"MGCY"
    || *(long*)ccd_pars.cfa_pattern == *(long*)"CYMG" )
   {
      pix_shift = true;
   }

   // write even rows in one field
   // odd rows in another field
   int half = image_H/2;
   for(int y=0; y<image_H; ++y)
   {
      int _y;
      if( y < half )
         _y = (y*2)+first;
      else
         _y = (y-half)*2+second;

      if( _y < 0 || _y >= ccd_pars.raw_height )
      {
         memset(row, 0, sizeof(row[0])*row_size );

         if( fwrite( row, row_size, 1, out ) != 1 )
         {
            ret = false;
            break;
         }
         continue;
      }

      color_t* pix = ccd + (ccd_pars.raw_width * _y);

      byte* r = row;

      if( pix_shift )
      {
         r[0] = (byte)(pix[1]>>4);
         r[1] = (byte)((pix[1]<<4)|((pix[0]>>8)&15));
         r[2] = (byte)(pix[0]);
         ++pix;
         r += 3;
         int end = ccd_pars.raw_width/2-1;
         for(int x=0; x<end; ++x, pix += 2, r += 3)
         {
            r[0] = (byte)(pix[0]>>4);
            r[1] = (byte)((pix[0]<<4)|((pix[1]>>8)&15));
            r[2] = (byte)(pix[1]);
         }
         if( !(ccd_pars.raw_width & 1) )
         {
            r[0] = (byte)(pix[0]>>4);
            r[1] = (byte)((pix[0]<<4)|((pix[-1]>>8)&15));
            r[2] = (byte)(pix[-1]);
         }
      }
      else
      {
         int end = ccd_pars.raw_width/2;
         for(int x=0; x<end; ++x, pix += 2, r += 3)
         {
            r[0] = (byte)(pix[0]>>4);
            r[1] = (byte)((pix[0]<<4)|((pix[1]>>8)&15));
            r[2] = (byte)(pix[1]);
         }
         if( ccd_pars.raw_width & 1 )
         {
            r[0] = (byte)(pix[0]>>4);
            r[1] = (byte)((pix[0]<<4)|((pix[-1]>>8)&15));
            r[2] = (byte)(pix[-1]);
         }
      }

      if( fwrite( row, row_size, 1, out ) != 1 )
      {
         ret = false;
         break;
      }
   }

   delete [] row;
   return ret;
}

////////////////////////////////////////////////////////////////////////////////

TCCDParam default_pars[] = {
   {1600+16, 1200+4, 0x2CE000, "YCGM", {{ 0, (1200+4)/2 }, {0x167000, (1200+4)/2}}  }, // E2500
   {2272+16, 1704+4, 0x598000, "GMYC", {{ 0, (1704+4)/2 }, {0x2CC000, (1704+4)/2}}  }, // E4500
   {2048+16, 1536+4, 0x48D000, "MGCY", {{ 0, (1536+4)/2 }, {0x246800, (1536+4)/2}}  }, // E995
};

const int n_default_pars = sizeof(default_pars) / sizeof(default_pars[0]);

void usage()
{
   printf("Coolpix RAW data converter to E5700 NEF file\n");
   printf(program " " version " [compiled: " datetime "], " author "\n");

   printf("\nargs: [options] input_file [output_file]\n");

   printf("\noptions: frame pars (for hex use 0x notation)\n");
   printf("   -w:1616 - raw width\n");
   printf("   -h:1204 - raw height\n");
   printf("   -c:YCGM - CFA pattern\n");
   printf("   -b0:0 - begin of field 0\n");
   printf("   -n0:602 - lines in field 0\n");
   printf("   -b1:167000 - begin of field 1\n");
   printf("   -n1:602 - lines in field 1\n");

   printf("\noptions: processing\n");
   printf("   -v - verbose\n");

   printf("\nSupported cameras: E2500, E4500, E995\n");
}

bool main(int n, char* a[])
{
   TCCDParam user_pars;
   memset( &user_pars, ~0, sizeof(user_pars));

   // endian: true for Intel, false for Motorolla
   bool endian = true;
   char* ifn = NULL;
   char* ofn = NULL;
   bool verbose = false;
   char ofn_buf[_MAX_PATH];

   int i;
   for(i=1; i<n; ++i)
   {
      if( a[i][0] == '-' )
      {
         switch( a[i][1] )
         {
         case 'c':
            memcpy(user_pars.cfa_pattern, &a[i][3], 4);
            break;

         case 'w':
            sscanf(&a[i][3], "%i", &user_pars.raw_width);
            break;

         case 'h':
            sscanf(&a[i][3], "%i", &user_pars.raw_height);
            break;

         case 'b': {
            int f = atoi(&a[i][2]); if( f >= 2 ) goto invalid;
            sscanf(&a[i][4], "%i", &user_pars.field[f].offset);
            } break;

         case 'n': {
            int f = atoi(&a[i][2]); if( f >= 2 ) goto invalid;
            sscanf(&a[i][4], "%i", &user_pars.field[f].rows);
            } break;

         case 'v': verbose = true; break;

         case '?':
            usage();
            return false;

         default:
            printf("unknown switch: %s\n", a[i]);
            break;

         invalid:
            printf("invalid switch: %s\n", a[i]);
            break;
         }
      }
      else
      {
         if( !ifn )
            ifn = a[i];
         else
            ofn = a[i];
      }
   }

   if( verbose )
      printf(program " " version " [compiled: " datetime "], " author "\n");

   if( !ifn )
   {
      printf("error: input file is missing\n\n");
      usage();
      return false;
   }
   if( !ofn )
   {
      char* fileName = ifn;
      char* ifn_slash = strrchr(fileName, '/');
      if( ifn_slash )
         fileName = ifn_slash;

      char* ifn_bslash = strrchr(fileName, '\\');
      if( ifn_bslash )
         fileName = ifn_bslash;

      char* ifn_ext = strrchr(fileName, '.');

      if( !ifn_ext )
         ifn_ext = ifn + strlen(ifn);

      int len = ifn_ext - ifn;

      ofn = ofn_buf;
      memcpy(ofn, ifn, len);
      strcpy(ofn+len, ".NEF");
   }

   TFile in = fopen( ifn, "rb");
   if( !in ) { printf("error: \"%s\" - can not open\n", ifn); return false; }

   fseek( in, 0, SEEK_END );
   unsigned file_size = ftell( in );

   if( verbose )
      printf("input: %s, size=%u\n", ifn, file_size);

   for(i=0; i<n_default_pars; ++i)
   {
      if( file_size == default_pars[i].file_size )
         break;
   }

   TCCDParam ccd_pars;

   if( i < n_default_pars )
   {
      if( verbose )
         printf("found default parameters for size=%u\n", file_size);

      ccd_pars = default_pars[i];
      if( user_pars.raw_width != ~0 ) ccd_pars.raw_width = user_pars.raw_width;
      if( user_pars.raw_height != ~0 ) ccd_pars.raw_height = user_pars.raw_height;
      if( user_pars.cfa_pattern[0] != -1 ) memcpy( ccd_pars.cfa_pattern, user_pars.cfa_pattern, 4);
      if( user_pars.field[0].offset != ~0 ) ccd_pars.field[0].offset = user_pars.field[0].offset;
      if( user_pars.field[0].rows != ~0 ) ccd_pars.field[0].rows = user_pars.field[0].rows;
      if( user_pars.field[1].offset != ~0 ) ccd_pars.field[1].offset = user_pars.field[1].offset;
      if( user_pars.field[1].rows != ~0 ) ccd_pars.field[1].rows = user_pars.field[1].rows;
   }
   else
   {
      if( verbose )
         printf("not found default parameters for size=%u\n", file_size);

      if( user_pars.raw_width == ~0 ) { printf("error: unknown raw width\n"); return false; }
      if( user_pars.raw_height == ~0 ) { printf("error: unknown raw height\n"); return false; }
      if( user_pars.cfa_pattern[0] == -1 ) { printf("error: unknown cfa_pattern\n"); return false; }
      if( user_pars.field[0].offset == ~0 ) { printf("error: unknown field 0 offset\n"); return false; }
      if( user_pars.field[0].rows == ~0 ) { printf("error: unknown field 0 rows\n"); return false; }
      if( user_pars.field[1].offset == ~0 ) { printf("error: unknown field 1 offset\n"); return false; }
      if( user_pars.field[1].rows == ~0 ) { printf("error: unknown field 1 rows\n"); return false; }

      ccd_pars = user_pars;
   }

   if( verbose )
   {
      printf("current parameters:\n");
      printf("raw width: %u\n", ccd_pars.raw_width);
      printf("raw height: %u\n", ccd_pars.raw_height);
      printf("cfa pattern: %4.4s\n", ccd_pars.cfa_pattern);
      printf("field 0: offset: 0x%08X rows: %u\n", ccd_pars.field[0].offset, ccd_pars.field[0].rows );
      printf("field 1: offset: 0x%08X rows: %u\n", ccd_pars.field[1].offset, ccd_pars.field[1].rows );
   }

   TBuff ccd(ccd_pars.raw_height*ccd_pars.raw_width);

   if( !ccd )
   {
      printf("error: memory allocation fault (%u)\n", ccd_pars.raw_height*ccd_pars.raw_width*sizeof(color_t));
      return false;
   }

   if( verbose )
      printf("read ccd data\n");

   if( !read_ccd( in, ccd, ccd_pars) )
   {
      printf("error: can not read ccd data\n");
      return false;
   }

   in.close();

   TFile out = fopen( ofn, "wb");
   if( !out ) { printf("error: \"%s\" - can not open\n", ofn); return false; }

   if( verbose )
      printf("write picture to: %s\n", ofn);

   if( !write_nef(out, endian, ccd, ccd_pars) )
      { printf("error: can not write NEF\n"); return false; }

   if( verbose )
      printf("OK\n");

   return true;
}

